import docker
import subprocess
import os
from simple_file_checksum import get_checksum
import zipfile
import requests

from rayasdk.constants import *
from rayasdk.messages import *
from rayasdk.logger import log, log_info, log_error


raya_imgs = {}
docker_client: docker.DockerClient = None

def __check_docker_instalation():
    global raya_imgs, docker_client
    # Check docker installation
    try:
        docker_client = docker.from_env()
    except docker.errors.DockerException:
        log_error(MSG_DOCKER_NOT_INSTALLER)
        # maybe required for MAC as well:
        if platform.system() == 'Windows':
            log_error(MSG_DOCKER_NOT_RUNNING)
        return False
    return True


def __remove_old_images(latest_version: str):   
    log_info(('Removing old Ra-Ya images ...'))
    for image in docker_client.images.list(all=True):
        if image.tags and RAYAENV_DOCKER_IMAGE_NAME in ":".join(image.tags):
            if f'{RAYAENV_DOCKER_IMAGE_NAME}:{latest_version}' not in image.tags: 
                docker_client.images.remove(image.id, force=True)
                log_info((f'\'{image.tags[0]}\' Image removed'))


def __download_image(version: str):
    log_info(f'Downloading Ra-Ya OS v{version} (only once)...')
    image_name = f'{RAYAENV_DOCKER_IMAGE_NAME}:{version}'
    try:
        if os.system(f'docker pull {image_name}') != 0:
            raise KeyboardInterrupt('')
    except KeyboardInterrupt:
        log_error('Failed to download Ra-Ya OS image. Please try again.')
        return False
    return True


def download_raya_image(version: str):
    if not __check_docker_instalation():
        return False
    __remove_old_images(version)
    if not __download_image(version=version):
        return False
    return True
    

def launch_command(output_file='', silent=False):
    if not os.path.exists(RAYAENV_DOCKER_UR_ROOT) or len(os.listdir(RAYAENV_DOCKER_UR_ROOT)) == 0:
        log_error(f'\'{RAYAENV_DOCKER_UR_ROOT}\' files does not exist')
        return False
    
    command = f'cd {RAYAENV_DOCKER_UR_ROOT}/scripts && ./run.container.sh {RAYAENV_DOCKER_VERSION}'
    docker_command = f'{command} '
    if silent:
        if output_file:
            docker_command += f'> {str(output_file)}'
        else:
            docker_command += f'> /dev/null'       
    else:
        if output_file:
            docker_command += f'| tee {str(output_file)}'
    return subprocess.Popen(docker_command, shell=True)


def launch_app_on_sim(app_id, args, debug=False):
    check_running_container()
    execution_route = f'/opt/raya_os/rayadevel/apps/{app_id}/'
    if debug:
        cmd = f'python3 -m debugpy --wait-for-client --listen 5678 __main__.py ' + ' '.join(args)
    else:
        cmd = f'python3 __main__.py ' + ' '.join(args)
    command = (
            'cd /opt/raya_os/rayadevel && '
            'source .bashrc && '
            f'cd {execution_route} && '
            f'{cmd}'
        )

    docker_command = (
            f'docker exec -it -u rayadevel {RAYAENV_DOCKER_CONTAINER_NAME} '
            f'bash -c "{command}"'
        )
    ret = subprocess.call(docker_command, shell=True)


def kill_apps():
    check_running_container()
    log_info(f'Killing all running apps...')
    command = 'pkill -9 -u rayadevel'
    docker_command = (
            f'docker exec -it {RAYAENV_DOCKER_CONTAINER_NAME} {command}'
        )
    ret = subprocess.call(docker_command, shell=True)


def check_container():
    
    response = requests.request("GET", f'{VCS_URL}')
    data = response.json()
    # Set parameters from VCS API
    docker_version = response['versionId']
    docker_url = response['link']
    docker_sha256 = response['checksum']
    
    
    # Check raya images
    for img in docker_client.images.list():
        tags = img.tags
        if tags:
            name, tag = tags[0].split(':', 1)
            if name == RAYAENV_DOCKER_IMGNAME:
                raya_imgs[tag] = img

    if RAYAENV_DOCKER_VERSION in raya_imgs:
        return True
    log_info(('Docker image '
              f'\'{RAYAENV_DOCKER_IMGNAME}:{docker_version}\' '
              'not found.'))

    if raya_imgs:
        log_info(('Removing old Ra-Ya containers and images...'))
        for container in docker_client.containers.list(all=True):
            tags = container.image.tags
            if tags:
                name, tag = tags[0].split(':', 1)
                if name == 'raya':
                    if container.status == 'running':
                        container.kill()
                    try:
                        container.remove()
                    except docker.errors.NotFound:
                        pass

        for img_tag in raya_imgs:
            docker_client.images.remove(
                image=f'{RAYAENV_DOCKER_IMGNAME}:{img_tag}'
            )

    log_info(f'Downloading Ra-Ya OS v{docker_version} (only once)...')
    try:
        os.system(f'gsutil -m cp -r {docker_url} {URSDK_TEMP_PATH}')
    except:
        log_error('Failed to download Ra-Ya OS image. Please try again.')
        return False

    log_info('Extracting Ra-Ya OS image.')
    with zipfile.ZipFile(docker_zipimgpath, 'r') as zip_ref:
        zip_ref.extractall(path=URSDK_TEMP_PATH)

    # TODO: implement verification of container
    """ 
    log_info(f'Checking downloaded file...')
    try:
        sha256 = get_checksum(RAYAENV_DOCKER_IMGPATH, algorithm="SHA256")
    except FileNotFoundError:
        log_error('Download error, try again.')
        return False
    if docker_sha256 != sha256:
        log_error('Verification Failed.')
        return False
    """

    log_info(f'Creating image...')
    ret = subprocess.call(
        f'docker load -i {str(docker_img_path)}',
        shell=True
    )
    if ret != 0:
        log_error('Error creating the image, try again.')
        return False

    return True


def check_running_container():
    # Check docker installation
    if not __check_docker_instalation():
        return False
    
    # Check running container
    container_running = False
    for container in docker_client.containers.list(all=True):
        if container.name == f'{RAYAENV_DOCKER_CONTAINER_NAME}':
            container_running = True

    if not container_running:
        raise Exception('The container is not running')
